import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from anyact import modulize

# fwd      : inputs↦outputs
# backprop : g_outputs, outputs, inputs ↦ g_inputs
def gradiator(fwd,backprop):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,*inputs):
			y = fwd(*inputs)
			ctx.save_for_backward(y,*inputs) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, go):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			y,*inputs = ctx.saved_tensors # restore input from context
			# print("bwd",y.shape,[a.shape for a in inputs])
			return backprop(go,y,*inputs)
	return GradiatorFunction.apply

def debug_gradiator():
	C = nn.Conv2d(1,1,3,padding=1)
	x = torch.randn(10,1,128,128)
	θ = torch.randn(10,1,128,128)
	y = torch.randn(10,1,128,128)
	G = gradiator(lambda x,θ:x+θ ,lambda g,y,x,θ:(x,θ))
	yy = G(C(x),θ)
	optim = torch.optim.Adam(C.parameters(),lr=1e-2)
	optim.zero_grad()
	nn.MSELoss()(y,yy).requires_grad_().backward()
	optim.step()


# forward  pass is computed with fun(x)
# backward pass is computed with grad_fun(x,fun(x),grad_out)
def gradiator_1(fun,grad_fun):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx, x):
			y = fun(x)
			ctx.save_for_backward(x,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_out):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,y = ctx.saved_tensors # restore input from context
			grad_in = grad_fun(x,y,grad_out)
			return grad_in
	return modulize(GradiatorFunction.apply)

# forward  pass is computed with fun(x,θ)
# backward pass is computed with grad_fun(x,θ,fun(x,θ),grad_out)
def gradiator_2(fun,grad_fun):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x,θ):
			y = fun(x,θ)
			ctx.save_for_backward(x,θ,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,θ,y = ctx.saved_tensors # restore input from context
			grad_x, grad_θ = grad_fun(x,θ,y,grad_o)
			return grad_x, grad_θ
	return GradiatorFunction.apply

# forward  pass is computed with fun(x,θ1,θ2)
# backward pass is computed with grad_fun(x,θ1,θ2,fun(x,θ1,θ2),grad_out)
def gradiator_3(fun,grad_fun):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x,θ1,θ2):
			y = fun(x,θ1,θ2)
			ctx.save_for_backward(x,θ1,θ2,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,θ1,θ2,y = ctx.saved_tensors # restore input from context
			grad_x, grad_θ1, grad_θ2 = grad_fun(x,θ1,θ2,y,grad_o)
			return grad_x, grad_θ1, grad_θ2
	return GradiatorFunction.apply
